"""Unit tests for parser and unparser modules."""

import unittest

import typed_ast.ast3
import typed_astunparse

from horast.ast_tools import ast_to_list
from horast.parser import parse
from horast.unparser import unparse
from .examples import EXAMPLES

MODE_RESULTS = {
    'exec': typed_ast.ast3.Module,
    #'eval': typed_ast.ast3.expr,
    'single': typed_ast.ast3.Interactive}

TYPE_COMMENT_EXAMPLES = [
    """a = 1""",
    """a = 1 # type: int""",
    """print('abc')""",
    """print('abc')\n# printing abc"""]

class Tests(unittest.TestCase):

    maxDiff = None

    def test_parser_mode(self):
        for mode, result in MODE_RESULTS.items():
            with self.subTest(mode=mode):
                tree = parse('print(1)\n# prints one', mode=mode)
                self.assertIsInstance(tree, result)

    def test_parse_type_comments(self):
        for example in TYPE_COMMENT_EXAMPLES:
            with self.subTest(nexample=example):
                tree = parse(example)
                self.assertIsNotNone(tree)
                code = unparse(tree)
                reparsed_tree = parse(code)
                self.assertIsNotNone(reparsed_tree)
                self.assertEqual(
                    typed_ast.ast3.dump(reparsed_tree), typed_ast.ast3.dump(tree),
                    '"""\n{}\n""" vs. original """\n{}\n"""'.format(code, example))
                self.assertEqual(code.strip(), example)

    def test_parse(self):
        for name, example in EXAMPLES.items():
            if ' with eol comments' in name or name.startswith('multiline '):
                continue
            with self.subTest(name=name, example=example):
                tree = parse(example)
                self.assertIsNotNone(tree)

    def test_parse_failure(self):
        with self.assertRaises(SyntaxError):
            parse('def ill_pass(): pass', mode='eval')

    def test_roundtrip_without_comments(self):
        for name, example in EXAMPLES.items():
            with self.subTest(name=name, example=example):
                tree = typed_ast.ast3.parse(example)
                code = unparse(tree)
                reparsed_tree = typed_ast.ast3.parse(code)
                self.assertEqual(typed_ast.ast3.dump(reparsed_tree), typed_ast.ast3.dump(tree))

    def test_roundtrip(self):
        only_localizable = False
        for name, example in EXAMPLES.items():
            if ' with eol comments' in name or name.startswith('multiline '):
                continue
            data = {}
            with self.subTest(name=name, example=example, data=data):
                tree = typed_ast.ast3.parse(example)
                code = typed_astunparse.unparse(tree)
                complete_tree = parse(example)
                data['complete_tree'] = complete_tree
                complete_code = unparse(complete_tree)
                data['complete_code'] = complete_code
                self.assertGreaterEqual(len(complete_code), len(code), (complete_code, code))
                reparsed_tree = typed_ast.ast3.parse(code)
                tree_nodes = ast_to_list(tree, only_localizable)
                reparsed_tree_nodes = ast_to_list(reparsed_tree, only_localizable)
                self.assertEqual(len(reparsed_tree_nodes), len(tree_nodes))
                self.assertEqual(typed_ast.ast3.dump(reparsed_tree), typed_ast.ast3.dump(tree))
                reparsed_complete_tree = parse(complete_code)
                complete_tree_nodes = ast_to_list(complete_tree, only_localizable)
                reparsed_complete_tree_nodes = ast_to_list(reparsed_complete_tree, only_localizable)
                self.assertEqual(
                    len(reparsed_complete_tree_nodes), len(complete_tree_nodes), complete_code)
                self.assertEqual(
                    typed_ast.ast3.dump(reparsed_complete_tree),
                    typed_ast.ast3.dump(complete_tree),
                    '"""\n{}\n""" vs. original """\n{}\n"""'.format(complete_code, example))



"""Various helper functions to query and manipulate AST."""

import ast
import logging
import re
import typing as t

import asttokens
from static_typing.ast_manipulation import RecursiveAstVisitor
import typed_ast.ast3

from .token_tools import Scope

_LOG = logging.getLogger(__name__)

AstPathNode = t.NamedTuple('AstPathNode', [
    ('node', typed_ast.ast3.AST), ('field', t.Optional[str]), ('index', t.Optional[int])])
"""Node on a AST path.

Meaning of fields:
- node is an AST node on the path
- field is name of the field in that node; None if node is the final (i.e. target) node of the path
- index is index of the next node in the field if value of the field is a list, None otherwise
"""


def ast_to_list(
        tree: typed_ast.ast3.AST, only_localizable: bool = False) -> t.List[typed_ast.ast3.AST]:
    """Generate a flat list of nodes in AST."""
    nodes = []
    class Visitor(RecursiveAstVisitor[typed_ast.ast3]):
        def visit_node(self, node):
            if not only_localizable or hasattr(node, 'lineno') and hasattr(node, 'col_offset'):
                nodes.append(node)
    visitor = Visitor()
    visitor.visit(tree)
    return nodes


def get_ast_node_locations(nodes: t.List[typed_ast.ast3.AST]) -> t.List[t.Tuple[int, int]]:
    return [(
        node.lineno if hasattr(node, 'lineno') else None,
        node.col_offset if hasattr(node, 'col_offset') else None) for node in nodes]


def convert_1d_str_index_to_2d(
        text: str, index: int, line_separator: str = '\n',
        newline_starts: t.List[int] = None) -> t.Tuple[int, int]:
    """Convert 1D index in text to tuple (lineno, col_offset)."""
    assert isinstance(text, str), type(text)
    assert isinstance(index, int), type(index)
    if newline_starts is None:
        newline_starts = [m.end() for m in re.finditer(line_separator, text)]
    if index < 0 or index > len(text):
        raise ValueError('index={} is outside [0,{}]'.format(index, len(text)))
    lineno = 1
    for newline_start in newline_starts:
        if newline_start > index:
            break
        lineno += 1
    col_offset = index
    if newline_starts and lineno > 1:
        col_offset -= newline_starts[lineno - 2]
    _LOG.debug('converted %s[%i] into (%i, %i)', repr(text), index, lineno, col_offset)
    return lineno, col_offset


def get_ast_node_scopes(code: str, nodes: t.List[typed_ast.ast3.AST]) -> t.List[Scope]:
    atok = asttokens.ASTTokens(code, tree=ast.parse(code))
    ast_nodes = ast_to_list(atok.tree)
    assert len(ast_nodes) == len(nodes), (len(ast_nodes), len(nodes))
    newline_starts = [m.end() for m in re.finditer('\n', code)]
    _LOG.debug('new lines in code: %s', newline_starts)
    scopes = []
    _LOG.debug('marked %i nodes:', len(ast_nodes))
    for node in ast_nodes:
        raw_node_scope = atok.get_text_range(node)
        _LOG.debug(
            'node %s at %s: code="""%s""", tree=%s',
            type(node).__name__, raw_node_scope, atok.get_text(node), ast.dump(node))
        node_scope = Scope(*[
            convert_1d_str_index_to_2d(code, _, newline_starts=newline_starts)
            for _ in raw_node_scope])
        _LOG.debug('scope %s is %s', raw_node_scope, node_scope)
        scopes.append(node_scope)
    return scopes


def node_path_in_ast(
        tree: typed_ast.ast3.AST,
        target_node: typed_ast.ast3.AST) -> t.List[AstPathNode]:
    """Find path to node in the given AST.

    Return a list of AstPathNode from root node up to the target node.
    """
    assert isinstance(tree, typed_ast.ast3.AST), type(tree)
    assert isinstance(target_node, typed_ast.ast3.AST), type(target_node)
    _LOG.debug('looking for node: %s', typed_ast.ast3.dump(target_node, include_attributes=True))
    nodes = ast_to_list(tree)
    nodes = nodes[:nodes.index(target_node) + 1]
    node_path = [AstPathNode(target_node, None, None)]
    current_anchor = target_node
    reversed_anchor_index = 0
    reversed_nodes = list(reversed(list(enumerate(nodes))))
    while current_anchor is not nodes[0]:
        reversed_anchor_index += 1
        for index, node in reversed_nodes[reversed_anchor_index:]:
            _LOG.debug('nodes[%i] is %s', index, node)
            for field_name, field_value in typed_ast.ast3.iter_fields(node):
                if field_value is None \
                        or isinstance(field_value, (int, float, str, type, tuple)) \
                        or isinstance(type(field_value), t.TypingMeta):
                    continue
                if isinstance(field_value, list):
                    found = False
                    for i, field_value_elem in enumerate(field_value):
                        if field_value_elem is current_anchor:
                            current_anchor = node
                            reversed_anchor_index += index
                            node_path.append(AstPathNode(node, field_name, i))
                            found = True
                            _LOG.debug('"%s[%i]" of %s is on the path', field_name, i, node)
                            break
                    if found:
                        break
                    continue
                if field_value is current_anchor:
                    current_anchor = node
                    reversed_anchor_index += index
                    node_path.append(AstPathNode(node, field_name, None))
                    _LOG.debug('"%s" of %s is on the path', field_name, node)
                    break
    return list(reversed(node_path))


def find_in_ast(
        code: str, tree: typed_ast.ast3.AST, nodes: t.List[typed_ast.ast3.AST],
        scope: Scope) -> t.Tuple[t.List[AstPathNode], bool]:
    """Return tuple: (path, before).

    Where:
    - path is path to the anchor node for the target scope
    - before is boolean flag set to True if target scope is before the anchor node, False otherwise
    """
    scopes = get_ast_node_scopes(code, nodes)
    assert len(nodes) == len(scopes), (len(nodes), len(scopes))
    node_scopes_by_start = list(zip(nodes, scopes))
    node_scopes_by_start.sort(key=lambda _: _[1].end, reverse=True)
    node_scopes_by_start.sort(key=lambda _: _[1].start)
    _LOG.debug('by start: %s', node_scopes_by_start)
    node_scopes_by_end = list(zip(nodes, scopes))
    node_scopes_by_end.sort(key=lambda _: _[1].start, reverse=True)
    node_scopes_by_end.sort(key=lambda _: _[1].end)
    _LOG.debug('by end: %s', node_scopes_by_end)
    target_scope = scope
    _LOG.debug('the target scope %s', target_scope)

    node_by_end_after_index = None
    for i, (node, scope) in enumerate(node_scopes_by_end):
        if target_scope.start > scope.end:
            node_by_end_after_index = i
            _LOG.debug('is after node %s at %s', node, scope)
            break
    if node_by_end_after_index is None:
        _LOG.debug('is not after any node')

    node_by_start_before_index = None
    for i, (node, scope) in enumerate(node_scopes_by_start):
        if target_scope.end < scope.start:
            node_by_start_before_index = i
            _LOG.debug('is before node %s at %s', node, scope)
            break
    if node_by_start_before_index is None:
        _LOG.debug('is not before any node')

    node_scopes_containing_target_scope = []
    for i, (node, scope) in enumerate(node_scopes_by_start):
        if scope.start > target_scope.start:
            break
        if scope.end > target_scope.end:
            node_scopes_containing_target_scope.append((node, scope))
    _LOG.debug('is within nodes %s', node_scopes_containing_target_scope)

    if node_by_end_after_index is None:
        _LOG.debug('target %s is before first node', target_scope)
        assert isinstance(nodes[0], (typed_ast.ast3.Module, typed_ast.ast3.Interactive)), type(nodes[0])
        return ([AstPathNode(nodes[0], 'body', 0)], True)

    if node_by_start_before_index is None and not node_scopes_containing_target_scope:
        _LOG.debug('target %s is after last node', target_scope)
        assert isinstance(nodes[0], (typed_ast.ast3.Module, typed_ast.ast3.Interactive)), type(nodes[0])
        assert len(nodes[0].body) > 0
        return ([AstPathNode(nodes[0], 'body', len(nodes[0].body) - 1)], False)
    elif node_by_start_before_index is None or not node_scopes_containing_target_scope:
        raise NotImplementedError(
            'inconsistent results for target {} in:\n"""\n{}\nafter {}, before {}, within {}"""'
            .format(target_scope, code, node_by_end_after_index, node_by_start_before_index,
                    node_scopes_containing_target_scope))

    assert node_by_end_after_index is not None
    assert node_by_start_before_index is not None
    assert len(node_scopes_containing_target_scope) > 0
    _LOG.debug(
        'target %s is neither before first node nor after last node in:\n"""\n%s\n"""'
        '\nbut between %s and %s, within %s',
        target_scope, code, node_scopes_by_end[node_by_end_after_index],
        node_scopes_by_start[node_by_start_before_index], node_scopes_containing_target_scope)
    within_node, _ = node_scopes_containing_target_scope[-1]
    before_node, _ = node_scopes_by_start[node_by_start_before_index]
    path = node_path_in_ast(tree, before_node)
    assert len(path) >= 2, path
    assert path[-2].node is within_node, (path[-2].node, within_node)
    return (path[:-1], True)


def insert_at_path_in_tree(
        tree: typed_ast.ast3.AST, inserted: typed_ast.ast3.AST,
        path_to_anchor: t.Sequence[AstPathNode],
        before_anchor: bool = False) -> typed_ast.ast3.AST:
    """Insert a new AST node into an existing AST at exactly specified location."""
    assert isinstance(tree, typed_ast.ast3.AST), type(tree)
    assert isinstance(inserted, typed_ast.ast3.AST), type(inserted)
    #assert isinstance(anchor, typed_ast.ast3.AST), type(anchor)
    parent, field, index = path_to_anchor[-1]
    if not before_anchor:
        index += 1
    getattr(parent, field).insert(index, inserted)
    return tree


def insert_in_tree(
        tree: typed_ast.ast3.AST, inserted: typed_ast.ast3.AST, anchor: typed_ast.ast3.AST,
        before_anchor: bool = False, strict: bool = False) -> typed_ast.ast3.AST:
    """Insert a new AST node into an existing AST near the anchor node.

    Try to maintain correctness after insertion.
    """
    assert isinstance(tree, typed_ast.ast3.AST), type(tree)
    assert isinstance(inserted, typed_ast.ast3.AST), type(inserted)
    assert isinstance(anchor, typed_ast.ast3.AST), type(anchor)
    node_path = node_path_in_ast(tree, anchor)
    if node_path is None:
        raise ValueError('the anchor node {} not found in AST {}'.format(anchor, tree))
    _LOG.debug('node path: %s', node_path)
    try:
        parent, field, index = node_path[-2]
    except IndexError as err:
        if strict:
            raise NotImplementedError('cannot insert at AST root') from err
        parent, field, index = node_path[0].node, 'body', 0
    _LOG.debug('parent of anchor node: %s', parent)
    if strict and index is None:
        raise NotImplementedError('cannot insert into a non-list field')
    node_path_index = -2
    if field == 'targets':
        index = None  # don't insert anything into assignment
    while index is None:
        node_path_index -= 1
        parent, field, index = node_path[node_path_index]
        if field == 'targets':
            index = None  # don't insert anything into assignment
    if before_anchor:
        pass
    else:
        index += 1
    getattr(parent, field).insert(index, inserted)
    return tree



"""Idenify docstrings in AST and transform the AST accordingly."""

import inspect

import typed_ast.ast3


def is_docstring(node: typed_ast.ast3.AST) -> bool:
    stack = inspect.stack()
    frame = stack[6].frame
    if 'fundef' not in frame.f_locals:
        return False
    expr = frame.f_locals['fundef'].body[0]
    if not isinstance(expr, typed_ast.ast3.Expr):
        return False
    if expr.value is not node:
        return False
    return True


def convert_docstrings(tree: typed_ast.ast3.AST) -> typed_ast.ast3.AST:
    raise NotImplementedError(tree)



"""Unit tests for ast_tools module."""

import unittest

import typed_ast.ast3

from horast.token_tools import Scope, get_comment_tokens, get_token_scopes
from horast.ast_tools import \
    ast_to_list, get_ast_node_locations, convert_1d_str_index_to_2d, get_ast_node_scopes, \
    find_in_ast
from .examples import EXAMPLES


class Tests(unittest.TestCase):

    def test_ast_to_list(self):
        for name, example in EXAMPLES.items():
            for only_localizable in (False, True):
                with self.subTest(name=name, example=example, only_localizable=only_localizable):
                    tree = typed_ast.ast3.parse(example)
                    nodes = ast_to_list(tree, only_localizable)
                    self.assertIsInstance(nodes, list)
                    if not only_localizable:
                        self.assertGreater(len(nodes), 0)

    def test_get_ast_node_locations(self):
        for name, example in EXAMPLES.items():
            for only_localizable in (False, True):
                with self.subTest(name=name, example=example, only_localizable=only_localizable):
                    tree = typed_ast.ast3.parse(example)
                    nodes = ast_to_list(tree, only_localizable)
                    locations = get_ast_node_locations(nodes)
                    self.assertIsInstance(locations, list)
                    for location in locations:
                        self.assertIsInstance(location, tuple)
                        self.assertEqual(len(location), 2)

    def test_convert_1d_str_index_to_2d(self):
        texts = [
            'def', 'def\n', 'def\nghi', '\ndef\nghi', 'abc\ndef\nghi',
            '', '\n', '\n\n']
        for text in texts:
            for index in range(len(text) + 1):
                with self.subTest(text=text, index=index):
                    location = convert_1d_str_index_to_2d(text, index)
                    lineno, col_offset = location
                    self.assertGreaterEqual(lineno, 1, location)
                    self.assertGreaterEqual(col_offset, 0, location)
                    lines = text.splitlines(keepends=True)
                    if not text or text.endswith('\n'):
                        lines += ['']
                    self.assertLessEqual(lineno, len(lines))
                    line = lines[lineno - 1]
                    self.assertLessEqual(
                        col_offset, len(line), 'line no.{} is {}'.format(lineno, repr(line)))
                    if col_offset < len(line):
                        self.assertEqual(line[col_offset], text[index])
                    else:
                        self.assertGreaterEqual(
                            lineno, len(lines),
                            'location {} oveflows into the next line in line no.{} {}'
                            .format(location, lineno, repr(line)))

    def test_get_ast_node_scopes(self):
        for name, example in EXAMPLES.items():
            with self.subTest(name=name, example=example):
                tree = typed_ast.ast3.parse(example)
                nodes = ast_to_list(tree)
                scopes = get_ast_node_scopes(example, nodes)
                self.assertIsInstance(scopes, list)
                self.assertEqual(len(scopes), len(nodes))
                for scope in scopes:
                    self.assertIsInstance(scope, Scope)
                locations = get_ast_node_locations(nodes)
                self.assertEqual(len(scopes), len(locations))
                for node, scope, location in zip(nodes, scopes, locations):
                    if None in location:
                        continue
                    self.assertEqual(scope.start, location, '{} in: """\n{}\n"""'.format(
                        typed_ast.ast3.dump(node), example))

    def test_find_in_ast(self):
        for name, example in EXAMPLES.items():
            if ' with eol comments' in name or name.startswith('multiline '):
                continue
            for index in range(0, len(example) + 1, 64):
                for end_index in range(index + 1, len(example) + 1, 64):
                    # TODO: decrease distance step without failing assertions
                    scope = Scope(convert_1d_str_index_to_2d(example, index),
                                  convert_1d_str_index_to_2d(example, end_index))
                    with self.subTest(name=name, example=example, scope=scope):
                        tree = typed_ast.ast3.parse(example)
                        nodes = ast_to_list(tree)
                        path, before = find_in_ast(example, tree, nodes, scope)
                        self.assertIsInstance(path, list)
                        self.assertIsInstance(before, bool)

    def test_find_in_ast_real(self):
        for name, example in EXAMPLES.items():
            if ' with eol comments' in name or name.startswith('multiline '):
                continue
            comment_tokens = get_comment_tokens(example)
            scopes = get_token_scopes(comment_tokens)
            for scope in scopes:
                with self.subTest(name=name, example=example, scope=scope):
                    tree = typed_ast.ast3.parse(example)
                    nodes = ast_to_list(tree)
                    path, before = find_in_ast(example, tree, nodes, scope)
                    self.assertIsInstance(path, list)
                    self.assertIsInstance(before, bool)



"""Extension of typed_ast parser for Python 3 that retains comments in the AST."""

import typed_ast.ast3

from .token_tools import get_comment_tokens
from .ast_comments import insert_comment_tokens


def parse(code: str, *args, **kwargs) -> typed_ast.ast3.AST:
    """Parse given code into AST based on typed_ast.ast3 with nodes as defined in horast.nodes."""
    assert isinstance(code, str), type(code)
    try:
        tree = typed_ast.ast3.parse(code, *args, **kwargs)
    except SyntaxError as err:
        raise SyntaxError('typed_ast.ast3.parse(code{}{}) failed on code:\n"""\n{}\n"""'.format(
            (', args=' + str(args)) if args else '', (', kwargs=' + str(kwargs)) if kwargs else '',
            code)) from err
    comment_tokens = get_comment_tokens(code)
    tree = insert_comment_tokens(code, tree, comment_tokens)
    return tree



"""Extension of typed_astunpare unparser for Python 3 that generates code with comments."""

import ast
import io
import logging

from astunparse.unparser import interleave
import typed_ast.ast3
import typed_astunparse

from .nodes import Comment

_LOG = logging.getLogger(__name__)


def interleave_noncomment(inter, f, seq):
    """Call f on each item in seq, calling inter() in between.
    """
    seq = iter(seq)
    prev = None
    try:
        prev = next(seq)
        f(prev)
    except StopIteration:
        pass
    else:
        for x in seq:
            if not isinstance(prev, Comment):
                inter()
            f(x)
            prev = x


class Unparser(typed_astunparse.Unparser):
    """Extension of typed_astunpare.Unparser that handles Comment nodes."""

    def _List(self, t):
        self.write("[")
        interleave_noncomment(lambda: self.write(", "), self.dispatch, t.elts)
        self.write("]")

    def _Set(self, t):
        assert(t.elts) # should be at least one element
        self.write("{")
        interleave_noncomment(lambda: self.write(", "), self.dispatch, t.elts)
        self.write("}")

    def _Dict(self, t):
        self.write("{")
        noncomment_keys = []
        comment_groups = []
        comment_group = []
        _LOG.info('keys  : %s', t.keys)
        _LOG.info('values: %s', t.values)
        for key in t.keys:
            if isinstance(key, Comment):
                comment_group.append(key)
                continue
            comment_groups.append(comment_group)
            comment_group = []
            noncomment_keys.append(key)
        noncomment_values = []
        comment_groups_iter = iter(comment_groups)
        try:
            comment_group = next(comment_groups_iter)
        except StopIteration:
            comment_group = None
        for value in t.values:
            if isinstance(value, Comment):
                comment_group.append(value)
                continue
            noncomment_values.append(value)
            try:
                comment_group = next(comment_groups_iter)
            except StopIteration:
                comment_group = None
        def write_triple(triple):
            (k, v, comments) = triple
            for comment in comments:
                self.dispatch(comment)
            self.dispatch(k)
            self.write(": ")
            self.dispatch(v)
        interleave(lambda: self.write(", "), write_triple, zip(
            noncomment_keys, noncomment_values, comment_groups))
        self.write("}")

    def _Tuple(self, t):
        self.write("(")
        if len(t.elts) == 1:
            (elt,) = t.elts
            self.dispatch(elt)
            self.write(",")
        else:
            interleave_noncomment(lambda: self.write(", "), self.dispatch, t.elts)
        self.write(")")

    def _Call(self, t):
        if isinstance(t, ast.Call):
            super()._Call(t)
            return

        self.dispatch(t.func)
        self.write("(")
        comma = False
        for e in t.args:
            if comma:
                self.write(", ")
            else:
                comma = True
            self.dispatch(e)
            if isinstance(e, Comment):
                comma = False
        for e in t.keywords:
            if comma:
                self.write(", ")
            else:
                comma = True
            self.dispatch(e)
            if isinstance(e, Comment):
                comma = False
        self.write(")")

    def _Comment(self, node):
        if node.eol:
            self.write('  #')
        else:
            self.fill('#')
        self.write(node.value.s)


def unparse(tree: typed_ast.ast3.AST, *args, **kwargs) -> str:
    """Unparse AST based on typed_ast.ast3 with nodes as defined in horast.nodes into code."""
    assert isinstance(tree, typed_ast.ast3.AST), type(tree)
    stream = io.StringIO()
    Unparser(tree, *args, file=stream, **kwargs)
    return stream.getvalue()



"""Unit tests for nodes module."""

import logging
import unittest

import typed_ast.ast3

from horast.nodes import Comment

_LOG = logging.getLogger(__name__)


class Tests(unittest.TestCase):

    @unittest.skip('eol field is temporarily disabled')
    def test_comment(self):
        expr_fields = typed_ast.ast3.Expr._fields
        comment_fields = Comment._fields
        _LOG.warning('%s', expr_fields)
        _LOG.warning('%s', comment_fields)
        self.assertGreater(len(comment_fields), len(expr_fields))



"""Extract comments from untouched source code and insert them into the AST."""

import logging
import tokenize
import typing as t

import typed_ast.ast3
import typed_astunparse

from .nodes import Comment
from .token_tools import get_token_locations, get_token_scopes
from .ast_tools import \
    ast_to_list, get_ast_node_locations, find_in_ast, insert_at_path_in_tree, insert_in_tree

_LOG = logging.getLogger(__name__)

def insert_comment_tokens(
        code: str, tree: typed_ast.ast3.AST,
        comment_tokens: t.List[tokenize.TokenInfo]) -> typed_ast.ast3.AST:
    assert isinstance(tree, typed_ast.ast3.AST)
    assert isinstance(comment_tokens, list)
    nodes = ast_to_list(tree)
    scopes = get_token_scopes(comment_tokens)
    for comment_token, scope in zip(comment_tokens, scopes):
        comment = Comment.from_token(comment_token, eol=False)
        path_to_anchor, before_anchor = find_in_ast(code, tree, nodes, scope)
        _LOG.debug('inserting %s %s %s', comment, 'before' if before_anchor else 'after',
                   path_to_anchor[-1])
        tree = insert_at_path_in_tree(tree, comment, path_to_anchor, before_anchor)
    return tree

def insert_comment_tokens_approx(
        tree: typed_ast.ast3.AST, tokens: t.List[tokenize.TokenInfo]) -> typed_ast.ast3.AST:
    assert isinstance(tree, typed_ast.ast3.AST)
    assert isinstance(tokens, list)
    token_locations = get_token_locations(tokens)
    _LOG.debug('token locations: %s', token_locations)
    nodes = ast_to_list(tree, only_localizable=True)
    if not nodes and tokens:
        _LOG.debug('overwriting empty AST with simplest editable tree')
        tree = typed_ast.ast3.Module(body=[], type_ignores=[], lineno=1, col_offset=0)
        nodes = ast_to_list(tree, only_localizable=True)
    node_locations = get_ast_node_locations(nodes)
    _LOG.debug('node locations: %s', node_locations)
    node_locations_iter = enumerate(node_locations)
    token_insertion_indices = []
    tokens_eol_status = []
    for token_index, token_location in enumerate(token_locations):
        eol_comment_here = False
        try:
            node_index, node_location = next(node_locations_iter)
        except StopIteration:
            node_index = len(node_locations)
            node_location = None
        while node_location is not None:
            token_line, _ = token_location
            node_line, _ = node_location
            if node_line > token_line:
                break
            if node_line == token_line:
                eol_comment_here = True
                if node_index < len(node_locations) - 1:
                    next_node_line, _ = node_locations[node_index + 1]
                    if next_node_line == token_line:
                        eol_comment_here = False
                #if eol_comment_here:
                #    raise NotImplementedError(
                #        'code "{}" and comment "{}" in line {}'
                #        ' -- only whole line comments are currently supported'
                #        .format(typed_astunparse.unparse(nodes[node_index]).strip(),
                #                tokens[token_index].string, node_line))
            try:
                node_index, node_location = next(node_locations_iter)
            except StopIteration:
                node_index = len(node_locations)
                break
        tokens_eol_status.append(eol_comment_here)
        token_insertion_indices.append(node_index)
    _LOG.debug('token insertion indices: %s', token_insertion_indices)
    _LOG.debug('tree before insertion:\n"""\n%s\n"""', typed_astunparse.dump(tree))
    _LOG.debug('code before insertion:\n"""\n%s\n"""', typed_astunparse.unparse(tree).strip())
    for token_index, token_insertion_index in reversed(list(enumerate(token_insertion_indices))):
        token = tokens[token_index]
        eol = tokens_eol_status[token_index]
        comment = Comment.from_token(token, eol)
        if token_insertion_index == 0:
            anchor = nodes[token_insertion_index]
            before_anchor = True
        elif token_insertion_index == len(node_locations):
            anchor = nodes[-1]
            before_anchor = False
        else:
            anchor = nodes[token_insertion_index - 1]
            before_anchor = False
        _LOG.debug('inserting %s %s %s', comment, 'before' if before_anchor else 'after', anchor)
        tree = insert_in_tree(tree, comment, anchor=anchor, before_anchor=before_anchor)
    _LOG.debug('tree after insertion:\n"""\n%s\n"""', typed_astunparse.dump(tree))
    #_LOG.warning('code after insertion:\n"""\n%s\n"""', typed_astunparse.unparse(tree).strip())
    return tree



"""Various helper functions to query and manipulate tokens."""

import io
import tokenize
import typing as t

Scope = t.NamedTuple('Scope', [('start', t.Tuple[int, int]), ('end', t.Tuple[int, int])])


def get_tokens(code: str) -> t.List[tokenize.TokenInfo]:
    assert isinstance(code, str), type(code)
    code_bytes = code.encode() # type: bytes
    with io.BytesIO(code_bytes) as code_bytes_reader:
        tokenizer = tokenize.tokenize(code_bytes_reader.readline)
        tokens = [token for token in tokenizer]
    return tokens


def get_comment_tokens(code: str, ignore_type_comments: bool = True) -> t.List[tokenize.TokenInfo]:
    tokens = [token for token in get_tokens(code) if token.type is tokenize.COMMENT]
    if ignore_type_comments:
        tokens = [token for token in tokens if not token.string.startswith('# type:')]
    return tokens


def get_token_locations(tokens: t.List[tokenize.TokenInfo]) -> t.List[t.Tuple[int, int]]:
    return [token.start for token in tokens]


def get_token_scopes(tokens: t.List[tokenize.TokenInfo]) -> t.List[Scope]:
    return [Scope(token.start, token.end) for token in tokens]



"""Unit tests for ast_comments module."""

import tokenize
import unittest

from horast.token_tools import get_tokens, get_comment_tokens
from .examples import EXAMPLES


class Tests(unittest.TestCase):

    def test_get_tokens(self):
        for name, example in EXAMPLES.items():
            with self.subTest(name=name, example=example):
                tokens = get_tokens(example)
                self.assertIsInstance(tokens, list)
                for token in tokens:
                    self.assertIsInstance(token, tokenize.TokenInfo)

    def test_get_comment_tokens(self):
        for name, example in EXAMPLES.items():
            with self.subTest(name=name, example=example):
                tokens = get_comment_tokens(example)
                self.assertIsInstance(tokens, list)
                for token in tokens:
                    self.assertIsInstance(token, tokenize.TokenInfo)
                    self.assertEqual(token.type, tokenize.COMMENT)



"""Below code is generic boilerplate and normally should not be changed.

To avoid setup script boilerplate, create "setup.py" file with the following minimal contents
and modify them according to the specifics of your package.

See the implementation of setup_boilerplate.Package for default metadata values and available
options.
"""

import importlib
import pathlib
import shutil
import sys
import typing as t

import setuptools

__updated__ = '2017-09-04'

SETUP_TEMPLATE = '''"""Setup script."""

import setup_boilerplate


class Package(setup_boilerplate.Package):

    """Package metadata."""

    name = ''
    description = ''
    download_url = 'https://github.com/mbdevpl/...'
    classifiers = [
        'Development Status :: 1 - Planning',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only']
    keywords = []


if __name__ == '__main__':
    Package.setup()
'''

HERE = pathlib.Path(__file__).resolve().parent


def find_version(
        package_name: str, version_module_name: str = '_version',
        version_variable_name: str = 'VERSION') -> str:
    """Simulate behaviour of "from package_name._version import VERSION", and return VERSION."""
    version_module = importlib.import_module(
        '{}.{}'.format(package_name.replace('-', '_'), version_module_name))
    return getattr(version_module, version_variable_name)


def find_packages(root_directory: str = '.') -> t.List[str]:
    """Find packages to pack."""
    exclude = ['test', 'test.*'] if ('bdist_wheel' in sys.argv or 'bdist' in sys.argv) else []
    packages_list = setuptools.find_packages(root_directory, exclude=exclude)
    return packages_list


def parse_readme(readme_path: str = 'README.rst', encoding: str = 'utf-8') -> str:
    """Read contents of readme file (by default "README.rst") and return them."""
    with open(str(HERE.joinpath(readme_path)), encoding=encoding) as readme_file:
        desc = readme_file.read()
    return desc


def parse_requirements(
        requirements_path: str = 'requirements.txt') -> t.List[str]:
    """Read contents of requirements.txt file and return data from its relevant lines.

    Only non-empty and non-comment lines are relevant.
    """
    requirements = []
    with open(str(HERE.joinpath(requirements_path))) as reqs_file:
        for requirement in [line.strip() for line in reqs_file.read().splitlines()]:
            if not requirement or requirement.startswith('#'):
                continue
            requirements.append(requirement)
    return requirements


def partition_version_classifiers(
        classifiers: t.Sequence[str], version_prefix: str = 'Programming Language :: Python :: ',
        only_suffix: str = ' :: Only') -> t.Tuple[t.List[str], t.List[str]]:
    """Find version number classifiers in given list and partition them into 2 groups."""
    versions_min, versions_only = [], []
    for classifier in classifiers:
        version = classifier.replace(version_prefix, '')
        versions = versions_min
        if version.endswith(only_suffix):
            version = version.replace(only_suffix, '')
            versions = versions_only
        try:
            versions.append(tuple([int(_) for _ in version.split('.')]))
        except ValueError:
            pass
    return versions_min, versions_only


def find_required_python_version(
        classifiers: t.Sequence[str], version_prefix: str = 'Programming Language :: Python :: ',
        only_suffix: str = ' :: Only') -> t.Optional[str]:
    """Determine the minimum required Python version."""
    versions_min, versions_only = partition_version_classifiers(
        classifiers, version_prefix, only_suffix)
    if len(versions_only) > 1:
        raise ValueError(
            'more than one "{}" version encountered in {}'.format(only_suffix, versions_only))
    only_version = None
    if len(versions_only) == 1:
        only_version = versions_only[0]
        for version in versions_min:
            if version[:len(only_version)] != only_version:
                raise ValueError(
                    'the "{}" version {} is inconsistent with version {}'
                    .format(only_suffix, only_version, version))
    min_supported_version = None
    for version in versions_min:
        if min_supported_version is None or \
                (len(version) >= len(min_supported_version) and version < min_supported_version):
            min_supported_version = version
    if min_supported_version is None:
        if only_version is not None:
            return '.'.join([str(_) for _ in only_version])
    else:
        return '>=' + '.'.join([str(_) for _ in min_supported_version])
    return None


class Package:

    """Default metadata and behaviour for a Python package setup script."""

    root_directory = '.' # type: str
    """Root directory of the source code of the package, relative to the setup.py file location."""

    name = None # type: str
    description = None # type: str
    url = 'https://mbdevpl.github.io/' # type: str
    download_url = 'https://github.com/mbdevpl' # type: str
    author = 'Mateusz Bysiek' # type: str
    author_email = 'mb@mbdev.pl' # type: str
    # maintainer = None # type: str
    # maintainer_email = None # type: str
    license_str = 'Apache License 2.0' # type: str

    classifiers = [] # type: t.List[str]
    """List of valid project classifiers: https://pypi.python.org/pypi?:action=list_classifiers"""

    keywords = [] # type: t.List[str]
    package_data = {}
    exclude_package_data = {}

    extras_require = {} # type: t.Mapping[str, t.List[str]]
    """A dictionary containing entries of type 'some_feature': ['requirement1', 'requirement2']."""

    entry_points = {} # type: t.Mapping[str, t.List[str]]
    """A dictionary used to enable automatic creation of console scripts, gui scripts and plugins.

    Example entry:
    'console_scripts': ['script_name = package.subpackage:function']
    """

    test_suite = 'test' # type: str

    @classmethod
    def try_fields(cls, *names) -> t.Optional[t.Any]:
        for name in names:
            if hasattr(cls, name):
                return getattr(cls, name)

    @classmethod
    def clean(cls, build_directory_name: str = 'build') -> None:
        """Recursively delete build directory (by default "build") if it exists."""
        build_directory_path = pathlib.Path(HERE, build_directory_name)
        if build_directory_path.is_dir():
            shutil.rmtree(str(build_directory_path))

    @classmethod
    def setup(cls) -> None:
        """Run setuptools.setup() with correct arguments."""
        setuptools.setup(
            name=cls.name, version=find_version(cls.name), description=cls.description,
            long_description=parse_readme(), url=cls.url, download_url=cls.download_url,
            author=cls.author, author_email=cls.author_email,
            maintainer=cls.try_fields('maintainer', 'author'),
            maintainer_email=cls.try_fields('maintainer_email', 'author_email'),
            license=cls.license_str, classifiers=cls.classifiers, keywords=cls.keywords,
            packages=find_packages(cls.root_directory), package_dir={'': cls.root_directory},
            include_package_data=True,
            package_data=cls.package_data, exclude_package_data=cls.exclude_package_data,
            install_requires=parse_requirements(), extras_require=cls.extras_require,
            python_requires=find_required_python_version(cls.classifiers),
            entry_points=cls.entry_points, test_suite=cls.test_suite
            )



"""Tests for setup scripts."""

import importlib
import itertools
import os
import pathlib
import runpy
import subprocess
import sys
import tempfile
import typing as t
import unittest

__updated__ = '2017-09-25'


def run_program(*args, glob: bool = False):
    """Run subprocess with given args. Use path globbing for each arg that contains an asterisk."""
    if glob:
        cwd = pathlib.Path.cwd()
        args = tuple(itertools.chain.from_iterable(
            list(str(_.relative_to(cwd)) for _ in cwd.glob(arg)) if '*' in arg else [arg]
            for arg in args))
    process = subprocess.Popen(args)
    process.wait()
    if process.returncode != 0:
        raise AssertionError('execution of {} returned {}'.format(args, process.returncode))
    return process


def run_pip(*args, **kwargs):
    python_exec_name = pathlib.Path(sys.executable).name
    pip_exec_name = python_exec_name.replace('python', 'pip')
    run_program(pip_exec_name, *args, **kwargs)


def run_module(name: str, *args, run_name: str = '__main__') -> None:
    backup_sys_argv = sys.argv
    sys.argv = [name + '.py'] + list(args)
    runpy.run_module(name, run_name=run_name)
    sys.argv = backup_sys_argv


def import_module(name: str = 'setup') -> 'module':
    setup_module = importlib.import_module(name)
    return setup_module


def import_module_member(module_name: str, member_name: str) -> t.Any:
    module = import_module(module_name)
    return getattr(module, member_name)


#def import_module_members(module_name: str, member_names: t.Iterable[str]) -> t.List[t.Any]:
#    module = import_module(module_name)
#    return [getattr(module, member_name) for member_name in member_names]


CLASSIFIERS_LICENSES = (
    'License :: OSI Approved :: Python License (CNRI Python License)',
    'License :: OSI Approved :: Python Software Foundation License',
    'License :: Other/Proprietary License',
    'License :: Public Domain')

CLASSIFIERS_PYTHON_VERSIONS = tuple("""Programming Language :: Python
Programming Language :: Python :: 2
Programming Language :: Python :: 2.3
Programming Language :: Python :: 2.4
Programming Language :: Python :: 2.5
Programming Language :: Python :: 2.6
Programming Language :: Python :: 2.7
Programming Language :: Python :: 2 :: Only
Programming Language :: Python :: 3
Programming Language :: Python :: 3.0
Programming Language :: Python :: 3.1
Programming Language :: Python :: 3.2
Programming Language :: Python :: 3.3
Programming Language :: Python :: 3.4
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Programming Language :: Python :: 3.7
Programming Language :: Python :: 3 :: Only""".splitlines())

CLASSIFIERS_PYTHON_IMPLEMENTATIONS = tuple("""Programming Language :: Python :: Implementation
Programming Language :: Python :: Implementation :: CPython
Programming Language :: Python :: Implementation :: IronPython
Programming Language :: Python :: Implementation :: Jython
Programming Language :: Python :: Implementation :: MicroPython
Programming Language :: Python :: Implementation :: PyPy
Programming Language :: Python :: Implementation :: Stackless""".splitlines())

CLASSIFIERS_VARIOUS = (
    'Framework :: IPython',
    'Topic :: Scientific/Engineering',
    'Topic :: Sociology',
    'Topic :: Security :: Cryptography',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Software Development :: Version Control :: Git',
    'Topic :: System',
    'Topic :: Utilities')

CLASSIFIERS_LICENSES_TUPLES = tuple((_,) for _ in CLASSIFIERS_LICENSES) + ((),)

CLASSIFIERS_PYTHON_VERSIONS_COMBINATIONS = tuple((_,) for _ in CLASSIFIERS_PYTHON_VERSIONS)

CLASSIFIERS_PYTHON_IMPLEMENTATIONS_TUPLES = tuple((_,) for _ in CLASSIFIERS_PYTHON_IMPLEMENTATIONS)

#CLASSIFIERS_VARIOUS_PERMUTATIONS = tuple(itertools.chain.from_iterable(
#    itertools.permutations(..., n)
#    for n in range(...)
#    ))

CLASSIFIERS_VARIOUS_COMBINATIONS = tuple(itertools.combinations(
    CLASSIFIERS_VARIOUS, len(CLASSIFIERS_VARIOUS) - 1)) + (CLASSIFIERS_VARIOUS,)

ALL_CLASSIFIERS_VARIANTS = [
    licenses + versions + implementations + various
    for licenses in CLASSIFIERS_LICENSES_TUPLES
    for versions in CLASSIFIERS_PYTHON_VERSIONS_COMBINATIONS
    for implementations in CLASSIFIERS_PYTHON_IMPLEMENTATIONS_TUPLES
    for various in CLASSIFIERS_VARIOUS_COMBINATIONS]


def get_package_folder_name():
    """Attempt to guess the built package name."""
    cwd = pathlib.Path.cwd()
    directories = [
        path for path in cwd.iterdir() if pathlib.Path(cwd, path).is_dir() \
            and pathlib.Path(cwd, path, '__init__.py').is_file() \
            and path.name != 'test']
    assert len(directories) == 1, directories
    return directories[0].name


class UnitTests(unittest.TestCase):
    """Test basic functionalities of the setup boilerplate."""

    def test_find_version(self):
        find_version = import_module_member('setup_boilerplate', 'find_version')
        result = find_version(get_package_folder_name())
        self.assertIsInstance(result, str)

    def test_find_packages(self):
        find_packages = import_module_member('setup_boilerplate', 'find_packages')
        results = find_packages()
        self.assertIsInstance(results, list)
        for result in results:
            self.assertIsInstance(result, str)

    def test_parse_readme(self):
        parse_readme = import_module_member('setup_boilerplate', 'parse_readme')
        result = parse_readme()
        self.assertIsInstance(result, str)

    def test_parse_reqs(self):
        parse_requirements = import_module_member('setup_boilerplate', 'parse_requirements')
        results = parse_requirements()
        self.assertIsInstance(results, list)
        self.assertTrue(all(isinstance(result, str) for result in results), msg=results)

    def test_parse_reqs_empty(self):
        parse_requirements = import_module_member('setup_boilerplate', 'parse_requirements')
        reqs_file = tempfile.NamedTemporaryFile('w', delete=False)
        reqs_file.close()
        results = parse_requirements(reqs_file.name)
        self.assertIsInstance(results, list)
        self.assertEqual(len(results), 0)
        os.remove(reqs_file.name)

    def test_parse_reqs_comments(self):
        parse_requirements = import_module_member('setup_boilerplate', 'parse_requirements')
        reqs = ['# comment', 'numpy', '', '# another comment', 'scipy', '', '# one more comment']
        reqs_file = tempfile.NamedTemporaryFile('w', delete=False)
        for req in reqs:
            print(req, file=reqs_file)
        reqs_file.close()
        results = parse_requirements(reqs_file.name)
        self.assertIsInstance(results, list)
        self.assertGreater(len(results), 0)
        self.assertLess(len(results), len(reqs))
        os.remove(reqs_file.name)

    def test_find_req_py_ver(self):
        find_required_python_version = import_module_member(
            'setup_boilerplate', 'find_required_python_version')
        for variant in ALL_CLASSIFIERS_VARIANTS:
            with self.subTest(variant=variant):
                result = find_required_python_version(variant)
                if result is not None:
                    self.assertIsInstance(result, str)

    def test_find_req_py_ver_reversed(self):
        find_required_python_version = import_module_member(
            'setup_boilerplate', 'find_required_python_version')
        classifiers = [
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6']
        req = find_required_python_version(classifiers)
        self.assertEqual(req, '>=3.4')
        req = find_required_python_version(reversed(classifiers))
        self.assertEqual(req, '>=3.4')

    def test_find_req_py_ver_none(self):
        find_required_python_version = import_module_member(
            'setup_boilerplate', 'find_required_python_version')
        result = find_required_python_version([])
        self.assertIsNone(result)

    def test_find_req_py_ver_many_only(self):
        find_required_python_version = import_module_member(
            'setup_boilerplate', 'find_required_python_version')
        classifiers = [
            'Programming Language :: Python :: 2 :: Only',
            'Programming Language :: Python :: 3 :: Only']
        with self.assertRaises(ValueError):
            find_required_python_version(classifiers)

    def test_find_req_py_ver_conflict(self):
        find_required_python_version = import_module_member(
            'setup_boilerplate', 'find_required_python_version')
        classifier_variants = [
            ['Programming Language :: Python :: 2.7',
             'Programming Language :: Python :: 3 :: Only'],
            ['Programming Language :: Python :: 2 :: Only',
             'Programming Language :: Python :: 3.0']]
        for classifiers in classifier_variants:
            with self.assertRaises(ValueError):
                find_required_python_version(classifiers)

    def test_package(self):
        package = import_module_member('setup_boilerplate', 'Package')
        class Package(package): # pylint: disable=too-few-public-methods
            name = 'package name'
            description = 'package description'
        self.assertIsNone(Package.try_fields())
        self.assertEqual(Package.try_fields('name', 'description'), 'package name')
        self.assertEqual(Package.try_fields('bad_field', 'description'), 'package description')


class IntergrationTests(unittest.TestCase):
    """Test if the boilerplate can actually create a valid package."""

    pkg_name = get_package_folder_name()

    @unittest.skipUnless(os.environ.get('TEST_PACKAGING'), 'skipping packaging test')
    def test_build_binary(self):
        run_module('setup', 'bdist')
        self.assertTrue(os.path.isdir('dist'))

    @unittest.skipUnless(os.environ.get('TEST_PACKAGING'), 'skipping packaging test')
    def test_build_wheel(self):
        run_module('setup', 'bdist_wheel')
        self.assertTrue(os.path.isdir('dist'))

    @unittest.skipUnless(os.environ.get('TEST_PACKAGING'), 'skipping packaging test')
    def test_build_source(self):
        run_module('setup', 'sdist', '--formats=gztar,zip')
        self.assertTrue(os.path.isdir('dist'))

    @unittest.skipUnless(os.environ.get('TEST_PACKAGING'), 'skipping packaging test')
    def test_clean(self):
        run_module('setup', 'bdist')
        self.assertTrue(os.path.isdir('build'))
        package = import_module_member('setup_boilerplate', 'Package')
        package.clean()
        os.mkdir('build')
        package.clean()
        package.clean()
        self.assertFalse(os.path.isdir('build'))

    @unittest.skipUnless(os.environ.get('TEST_PACKAGING'), 'skipping packaging test')
    def test_install_code(self):
        run_pip('install', '.')
        run_pip('uninstall', '-y', self.pkg_name)

    @unittest.skipUnless(os.environ.get('TEST_PACKAGING'), 'skipping packaging test')
    def test_install_source_tar(self):
        find_version = import_module_member('setup_boilerplate', 'find_version')
        version = find_version(self.pkg_name)
        run_pip('install', 'dist/*-{}.tar.gz'.format(version), glob=True)
        run_pip('uninstall', '-y', self.pkg_name)

    @unittest.skipUnless(os.environ.get('TEST_PACKAGING'), 'skipping packaging test')
    def test_install_source_zip(self):
        find_version = import_module_member('setup_boilerplate', 'find_version')
        version = find_version(self.pkg_name)
        run_pip('install', 'dist/*-{}.zip'.format(version), glob=True)
        run_pip('uninstall', '-y', self.pkg_name)

    @unittest.skipUnless(os.environ.get('TEST_PACKAGING'), 'skipping packaging test')
    def test_install_wheel(self):
        find_version = import_module_member('setup_boilerplate', 'find_version')
        version = find_version(self.pkg_name)
        run_pip('install', 'dist/*-{}-*.whl'.format(version), glob=True)
        run_pip('uninstall', '-y', self.pkg_name)

    def test_pip_error(self):
        with self.assertRaises(AssertionError):
            run_pip('wrong_pip_command')

    def test_setup_do_nothing_or_error(self):
        run_module('setup', 'wrong_setup_command', run_name='__not_main__')
        with self.assertRaises(SystemExit):
            run_module('setup', 'wrong_setup_command')



"""Definitions of AST nodes used to store missing syntax information."""

import tokenize
import typing as t

import typed_ast.ast3


def _comment_token_to_ast_node(token: tokenize.TokenInfo) -> typed_ast.ast3.Str:
    """Convert comment token to its raw AST representation."""
    return typed_ast.ast3.Str(s=token.string[1:], lineno=token.start[0], col_offset=token.start[1])


class Comment(typed_ast.ast3.Expr):
    """Store code comment in AST.

    Examples:

    # full line comment

    expr # end-of-line comment

    (value1,
     # full line comment inside an expression
     value2,
     value3)

    (value1,
     value2,  # end-of line comment inside an expression
     value3)
    """

    #_fields = typed_ast.ast3.Expr._fields + ('eol',)

    @classmethod
    def from_token(cls, token: tokenize.TokenInfo, eol: bool):
        return cls(
            value=_comment_token_to_ast_node(token), eol=eol,
            lineno=token.start[0], col_offset=token.start[1])


class BlockComment(typed_ast.ast3.Expr):
    """Store sequence of code comments as a single node in AST.

    Example:

    # 1st line of a comment
    # 2nd line of a comment
    # 3rd line of a comment
    """

    @classmethod
    def from_tokens(cls, tokens: t.List[tokenize.TokenInfo]):
        return cls(
            value=typed_ast.ast3.Tuple(
                elts=[_comment_token_to_ast_node(token) for token in tokens],
                lineno=tokens[0].start[0], col_offset=tokens[0].start[1]),
            lineno=tokens[0].start[0], col_offset=tokens[0].start[1])


class Docstring(typed_ast.ast3.Expr):
    """Store docstring as a special node in AST."""

    @classmethod
    def from_str_expr(cls, str_expr: typed_ast.ast3.Expr):
        raise NotImplementedError()



"""Unit tests for ast_comments module."""

import unittest

import typed_ast.ast3

from horast.token_tools import get_comment_tokens
from horast.ast_tools import ast_to_list
from horast.ast_comments import insert_comment_tokens, insert_comment_tokens_approx
from .examples import EXAMPLES


class Tests(unittest.TestCase):

    def test_comment_tokens(self):
        only_localizable = False
        for name, example in EXAMPLES.items():
            if ' with eol comments' in name or name.startswith('multiline '):
                continue
            with self.subTest(name=name, example=example):
                tree = insert_comment_tokens(
                    example, typed_ast.ast3.parse(example), get_comment_tokens(example))
                tree_nodes = ast_to_list(tree, only_localizable)
                self.assertIsInstance(tree, typed_ast.ast3.AST)
                nodes = ast_to_list(typed_ast.ast3.parse(example), only_localizable)
                comments = get_comment_tokens(example)
                self.assertEqual(
                    len(tree_nodes), max(1 if comments else 0, len(nodes)) + 2 * len(comments),
                    (tree_nodes, nodes, comments))

    def test_comment_tokens_approx(self):
        for name, example in EXAMPLES.items():
            for only_localizable in (False, True):
                with self.subTest(name=name, example=example, only_localizable=only_localizable):
                    tree = insert_comment_tokens_approx(
                        typed_ast.ast3.parse(example), get_comment_tokens(example))
                    tree_nodes = ast_to_list(tree, only_localizable)
                    self.assertIsInstance(tree, typed_ast.ast3.AST)
                    nodes = ast_to_list(typed_ast.ast3.parse(example), only_localizable)
                    comments = get_comment_tokens(example)
                    self.assertEqual(
                        len(tree_nodes), max(1 if comments else 0, len(nodes)) + 2 * len(comments),
                        (tree_nodes, nodes, comments))
